@@ -24,11 +24,12 @@ local exports = {}
 local E = {}
 local N = "lua_redis"
 
+local db_schema = (ts.number / tostring + ts.string):is_optional():describe("Database number")
 local common_schema = {
   timeout = (ts.number + ts.string / lutil.parse_time_interval):is_optional():describe("Connection timeout"),
-  db = ts.string:is_optional():describe("Database number"),
-  database = ts.string:is_optional():describe("Database number"),
-  dbname = ts.string:is_optional():describe("Database number"),
+  db = db_schema,
+  database = db_schema,
+  dbname = db_schema,
   prefix = ts.string:is_optional():describe("Key prefix"),
   username = ts.string:is_optional():describe("Username"),
   password = ts.string:is_optional():describe("Password"),
@@ -64,7 +65,8 @@ local server_schema = lutil.table_merge({
 
 local enrich_schema = function(external)
   return ts.one_of {
-    ts.shape(external), -- no specific redis parameters
+    ts.shape(lutil.table_merge(common_schema,
+        external)), -- no specific redis servers (e.g when global settings are used)
     ts.shape(lutil.table_merge(read_schema, external)), -- read_servers specified
     ts.shape(lutil.table_merge(write_schema, external)), -- write_servers specified
     ts.shape(lutil.table_merge(rw_schema, external)), -- both read and write servers defined
@@ -1138,6 +1140,8 @@ local function script_set_loaded(script)
     script.loaded = true
   end
 
+  script.pending_upload = false -- Allow further reload
+
   local wait_table = {}
   for _, s in ipairs(script.waitq) do
     table.insert(wait_table, s)
@@ -1162,8 +1166,11 @@ local function prepare_redis_call(script)
   end
 
   -- Call load script on each server, set loaded flag
-  script.in_flight = #servers
-  for _, s in ipairs(servers) do
+  if not script.servers_ready then
+    script.servers_ready = { } -- possible values for each server: unsent, tempfail, failed, done
+  end
+
+  for idx, s in ipairs(servers) do
     local cur_opts = {
       host = s:get_addr(),
       timeout = script.redis_params['timeout'],
@@ -1172,6 +1179,11 @@ local function prepare_redis_call(script)
       upstream = s
     }
 
+    -- By default we start from unsent status
+    if not script.servers_ready[idx] then
+      script.servers_ready[idx] = "unsent"
+    end
+
     if script.redis_params['username'] then
       cur_opts['username'] = script.redis_params['username']
     end
@@ -1190,46 +1202,91 @@ local function prepare_redis_call(script)
   return options
 end
 
+local function is_all_servers_ready(script)
+  for _, s in ipairs(script.servers_ready) do
+    if s == "unsent" or s == "tempfail" then
+      return false
+    end
+  end
+
+  -- We assume that permanent errors are not recoverable, so we will just skip those servers
+  return true
+end
+
+local function is_all_servers_failed(script)
+  for _, s in ipairs(script.servers_ready) do
+    if s == "unsent" or s == "tempfail" or s == "done" then
+      return false
+    end
+  end
+
+  return true
+end
+
+local function script_description(script)
+  return script.filename and ("from file: " .. script.filename)
+      or ("with id: " .. script.id)
+end
+
 local function load_script_task(script, task, is_write)
   local rspamd_redis = require "rspamd_redis"
   local opts = prepare_redis_call(script)
 
-  for _, opt in ipairs(opts) do
-    opt.task = task
-    opt.is_write = is_write
-    opt.callback = function(err, data)
-      if err then
-        logger.errx(task, 'cannot upload script to %s: %s; registered from: %s:%s',
-            opt.upstream:get_addr():to_string(true),
-            err, script.caller.short_src, script.caller.currentline)
+  for idx, opt in ipairs(opts) do
+    if script.servers_ready[idx] ~= 'done' then
+      opt.task = task
+      opt.is_write = is_write
+      opt.callback = function(err, data)
+        if err then
+          if string.match(err, 'ERR') then
+            logger.errx(task, 'cannot upload script %s to %s: %s; registered from: %s:%s',
+                script_description(script),
+                opt.upstream:get_addr():to_string(true),
+                err, script.caller.short_src, script.caller.currentline)
+            opt.upstream:fail()
+            script.servers_ready[idx] = "failed"
+            return
+          else
+            -- Assume temporary error
+            logger.infox(task, 'temporary error uploading script %s to %s: %s; registered from: %s:%s',
+                script_description(script),
+                opt.upstream:get_addr():to_string(true),
+                err, script.caller.short_src, script.caller.currentline)
+            script.servers_ready[idx] = "tempfail"
+            return
+          end
+        else
+          opt.upstream:ok()
+          logger.infox(task,
+              "uploaded redis script to %s %s, sha: %s",
+              opt.upstream:get_addr():to_string(true),
+              script_description(script), data)
+          script.sha = data -- We assume that sha is the same on all servers
+          script.servers_ready[idx] = "done"
+        end
+        if is_all_servers_ready(script) then
+          script_set_loaded(script)
+        elseif is_all_servers_failed(script) then
+          script.pending_upload = false
+          script.fatal_error = "cannot upload script to any server"
+        end
+      end -- callback
+
+      local ret = rspamd_redis.make_request(opt)
+
+      if not ret then
+        logger.errx('cannot execute redis request to load script on %s',
+            opt.upstream:get_addr())
+        script.servers_ready[idx] = "failed"
         opt.upstream:fail()
-        script.fatal_error = err
-      else
-        opt.upstream:ok()
-        logger.infox(task,
-            "uploaded redis script to %s %s %s, sha: %s",
-            opt.upstream:get_addr():to_string(true),
-            script.filename and "from file" or "with id", script.filename or script.id, data)
-        script.sha = data -- We assume that sha is the same on all servers
-      end
-      script.in_flight = script.in_flight - 1
-
-      if script.in_flight == 0 then
-        script_set_loaded(script)
       end
     end
 
-    local ret = rspamd_redis.make_request(opt)
-
-    if not ret then
-      logger.errx('cannot execute redis request to load script on %s',
-          opt.upstream:get_addr())
-      script.in_flight = script.in_flight - 1
-      opt.upstream:fail()
-    end
-
-    if script.in_flight == 0 then
+    if is_all_servers_ready(script) then
       script_set_loaded(script)
+    elseif is_all_servers_failed(script) then
+      script.pending_upload = false
+      script.fatal_error = "cannot upload script to any server"
     end
   end
 end
@@ -1238,44 +1295,64 @@ local function load_script_taskless(script, cfg, ev_base, is_write)
   local rspamd_redis = require "rspamd_redis"
   local opts = prepare_redis_call(script)
 
-  for _, opt in ipairs(opts) do
-    opt.config = cfg
-    opt.ev_base = ev_base
-    opt.is_write = is_write
-    opt.callback = function(err, data)
-      if err then
-        logger.errx(cfg, 'cannot upload script to %s: %s; registered from: %s:%s, filename: %s',
-            opt.upstream:get_addr():to_string(true),
-            err, script.caller.short_src, script.caller.currentline, script.filename)
+  for idx, opt in ipairs(opts) do
+    if script.servers_ready[idx] ~= 'done' then
+      opt.config = cfg
+      opt.ev_base = ev_base
+      opt.is_write = is_write
+      opt.callback = function(err, data)
+        if err then
+          if string.match(err, 'ERR') then
+            logger.errx(cfg, 'cannot upload script %s to %s: %s; registered from: %s:%s',
+                script_description(script),
+                opt.upstream:get_addr():to_string(true),
+                err, script.caller.short_src, script.caller.currentline)
+            opt.upstream:fail()
+            script.servers_ready[idx] = "failed"
+            return
+          else
+            -- Assume temporary error
+            logger.infox(cfg, 'temporary error uploading script %s to %s: %s; registered from: %s:%s',
+                script_description(script),
+                opt.upstream:get_addr():to_string(true),
+                err, script.caller.short_src, script.caller.currentline)
+            script.servers_ready[idx] = "tempfail"
+            return
+          end
+        else
+          opt.upstream:ok()
+          logger.infox(cfg,
+              "uploaded redis script %s to %s, sha: %s",
+              script_description(script),
+              opt.upstream:get_addr():to_string(true),
+              data)
+          script.sha = data -- We assume that sha is the same on all servers
+          script.servers_ready[idx] = "done"
+        end
+
+        if is_all_servers_ready(script) then
+          script_set_loaded(script)
+        elseif is_all_servers_failed(script) then
+          script.pending_upload = false
+          script.fatal_error = "cannot upload script to any server"
+        end
+      end
+      local ret = rspamd_redis.make_request(opt)
+
+      if not ret then
+        logger.errx('cannot execute redis request to load script %s on %s',
+            script_description(script),
+            opt.upstream:get_addr())
+        script.servers_ready[idx] = "failed"
         opt.upstream:fail()
-        script.fatal_error = err
-      else
-        opt.upstream:ok()
-        logger.infox(cfg,
-            "uploaded redis script to %s %s %s, sha: %s",
-            opt.upstream:get_addr():to_string(true),
-            script.filename and "from file" or "with id", script.filename or script.id,
-            data)
-        script.sha = data -- We assume that sha is the same on all servers
-        script.fatal_error = nil
-      end
-      script.in_flight = script.in_flight - 1
-
-      if script.in_flight == 0 then
-        script_set_loaded(script)
       end
     end
-    local ret = rspamd_redis.make_request(opt)
 
-    if not ret then
-      logger.errx('cannot execute redis request to load script on %s',
-          opt.upstream:get_addr())
-      script.in_flight = script.in_flight - 1
-      opt.upstream:fail()
-    end
-
-    if script.in_flight == 0 then
+    if is_all_servers_ready(script) then
       script_set_loaded(script)
+    elseif is_all_servers_failed(script) then
+      script.pending_upload = false
+      script.fatal_error = "cannot upload script " .. script_description(script) .. " to any server"
     end
   end
 end
@@ -1306,12 +1383,17 @@ local function add_redis_script(script, redis_params, caller_level, maybe_filena
   rspamd_config:add_on_load(function(cfg, ev_base, worker)
     local mult = 0.0
     rspamd_config:add_periodic(ev_base, 0.0, function()
-      if not new_script.sha then
+      if not new_script.sha and not new_script.fatal_error then
         load_redis_script(new_script, cfg, ev_base, worker)
-        mult = mult + 1
+
+        -- Do not wait for more than 10 seconds to retry
+        if mult < 10 then
+          mult = mult + 1
+        end
         return 1.0 * mult -- Check one more time in one second
       end
 
+      -- All loaded, we are good to go
       return false
     end, false)
   end)
@@ -1385,17 +1467,26 @@ local function exec_redis_script(id, params, callback, keys, args)
       if not err then
         callback(err, data)
       elseif string.match(err, 'NOSCRIPT') then
-        -- Schedule restart
-        script.sha = nil
+        -- Schedule restart if possible
         if can_reload then
           table.insert(script.waitq, do_call)
-          if script.in_flight == 0 then
+          if not script.pending_upload then
+            logger.infox(params.task or rspamd_config,
+                'redis script %s is not loaded (NOSCRIPT returned), scheduling reload',
+                script_description(script))
+            script.sha = nil
+            script.loaded = nil
+            script.pending_upload = true
             -- Reload scripts if this has not been initiated yet
             if params.task then
               load_script_task(script, params.task)
             else
               load_script_taskless(script, rspamd_config, params.ev_base)
             end
+          else
+            logger.infox(params.task or rspamd_config,
+                'redis script %s is not ready (NOSCRIPT returned), waiting it to be loaded',
+                script_description(script))
           end
         else
           callback(err, data)
@@ -1433,22 +1524,24 @@ local function exec_redis_script(id, params, callback, keys, args)
     end
   end
 
-  if script.loaded then
+  if script.loaded and script.sha then
     do_call(true)
   else
-    -- Delayed until scripts are loaded
-    if not params.task then
-      table.insert(script.waitq, do_call)
+    table.insert(script.waitq, do_call)
+    if not script.servers_ready then
+      -- Delayed until scripts are loaded
+      logger.infox(params.task or rspamd_config, 'redis script %s is not loaded, trying to load',
+          script_description(script))
+      -- Reload scripts if this has not been initiated yet
+      if params.task then
+        load_script_task(script, params.task)
+      else
+        load_script_taskless(script, rspamd_config, params.ev_base)
+      end
     else
-      -- TODO: fix taskfull requests
-      table.insert(script.waitq, function()
-        if script.loaded then
-          do_call(false)
-        else
-          callback('NOSCRIPT', nil)
-        end
-      end)
-      load_script_task(script, params.task, params.is_write)
+      -- Already loaded, but not yet ready
+      logger.infox(params.task or rspamd_config, 'redis script %s is not ready, waiting it to be loaded',
+          script_description(script))
     end
   end
 
@@ -1645,11 +1738,10 @@ exports.request = function(redis_params, attrs, req)
     opts.dbname = redis_params.db
   end
 
-  lutil.debugm(N, 'perform generic request to redis server' ..
-      ' (host=%s, timeout=%s): cmd: %s, arguments: %s', addr,
-      opts.timeout, opts.cmd, opts.args)
-
   if opts.callback then
+    lutil.debugm(N, 'perform generic async request to redis server' ..
+        ' (host=%s, timeout=%s): cmd: %s, arguments: %s', addr,
+        opts.timeout, opts.cmd, opts.args)
     local ret, conn = rspamd_redis.make_request(opts)
     if not ret then
       logger.errx(log_obj, 'cannot execute redis request')
@@ -1659,6 +1751,9 @@ exports.request = function(redis_params, attrs, req)
     return ret, conn, addr
   else
     -- Coroutines version
+    lutil.debugm(N, 'perform generic coroutine request to redis server' ..
+        ' (host=%s, timeout=%s): cmd: %s, arguments: %s', addr,
+        opts.timeout, opts.cmd, opts.args)
     local ret, conn = rspamd_redis.connect_sync(opts)
     if not ret then
       logger.errx(log_obj, 'cannot execute redis request')
